using System;
using System.IO;
using System.Collections;
using System.Text;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.util.encoders;
using org.bouncycastle.util.test;

namespace test.crypto
{
	/**
	* Test vectors from the NIST standard tests and Brian Gladman's vector set
	* <a href="http://fp.gladman.plus.com/cryptography_technology/rijndael/">
	* http://fp.gladman.plus.com/cryptography_technology/rijndael/</a>
	*/
	public class AESVectorFileTest: Test
	{

		private int countOfTests = 0;
		private int testNum = 0;

		protected virtual BlockCipher createNewEngineForTest() {
			return new AESEngine();
		}
	        
		private Test[] readTestVectors(Stream inStream) 
		{
			// initialize key, plaintext, ciphertext = null
			// read until find BLOCKSIZE=
			// return if not 128
			// read KEYSIZE= or ignore
			// loop
			//  read a line
			// if starts with BLOCKSIZE=
			//    parse the rest. return if not 128
			// if starts with KEY=
			//    parse the rest and set KEY
			// if starts with PT=
			//    parse the rest and set plaintext
			// if starts with CT=
			//    parse the rest and set ciphertext
			// if starts with TEST= or end of file
			//       if key, plaintext, ciphertext are all not null
			//          save away their values as the next test
			// until end of file
			ArrayList tests = new ArrayList();
			String key = null;
			String plaintext = null;
			String ciphertext = null;

			StreamReader input = new StreamReader(inStream);

			try {
				for (String line = input.ReadLine(); (line != null) ;) {
					line = line.Trim().ToLower();
					if (line.StartsWith("blocksize=")) {
						int i=0;
						try {
							i = Int32.Parse(line.Substring(10).Trim());
						} catch (Exception e) {
							Console.WriteLine("Error in paring blocksize: " + e);
						}
						if (i != 128) {
							return null;
						}
					} else if (line.StartsWith("keysize=")) {
						int i=0;
						try {
							i = Int32.Parse(line.Substring(10).Trim());
						} catch (Exception e) {
							Console.WriteLine("Error in paring keysize: " + e);
						}
						if ((i != 128) && (i != 192) && (i != 256)) {
							return null;
						}
					} else if (line.StartsWith("key=")) {
						key = line.Substring(4).Trim();
					} else if (line.StartsWith("pt=")) {
						plaintext = line.Substring(3).Trim();
					} else if (line.StartsWith("ct=")) {
						ciphertext = line.Substring(3).Trim();
					} else if (line.StartsWith("test=")) {
						if ((key != null) && (plaintext != null) && (ciphertext != null)) {
							tests.Add(new BlockCipherVectorTest
								(testNum++, 
								 createNewEngineForTest(),
								 new KeyParameter(Hex.decode(key)), plaintext, ciphertext));
						}
					}
	                
					line = input.ReadLine();
				}
				try {
					input.Close();
				} catch (IOException e) {
					Console.WriteLine("Error in closing stream: " + e);
				}
			} catch (IOException e) {
				Console.WriteLine("Error in parsing input files: " + e);
			}
			if ((key != null) && (plaintext != null) && (ciphertext != null)) {
				tests.Add(new BlockCipherVectorTest(testNum++, 
													createNewEngineForTest(),
													new KeyParameter(Hex.decode(key)), 
													plaintext, 
													ciphertext));
			}
			return (Test[])(tests.ToArray(typeof(Test)));
		}


		public virtual String getName() {
			return "AES";
		}


		private TestResult performTestsFromZipFile(String zfile) 
		{
			try {
				for (int i = 44; i <= 88;)
				{
					Test[] tests = null;

					try {
						String infile = "data\\" + zfile + i.ToString() + ".txt";
						FileStream stream = File.Open (infile, FileMode.Open,
							FileAccess.Read);
						tests = readTestVectors(stream);
					} 
					catch (Exception e) 
					{
						return new SimpleTestResult(false, getName()+": threw "+e);
					}
					if (tests != null) {
						for (int j = 0; j != tests.Length; j++) {
							TestResult  res = tests[j].perform();
							countOfTests++;
	                        
							if (!res.isSuccessful()) {
								return res;
							}
						}
					}
					if (i == 48 || i == 58 || i == 68 || i == 78) 
						i += 6;
					else
						i++;
					
				}

				return new SimpleTestResult(true, getName() + ": Okay");
			} 
			catch (Exception e) 
			{
				return new SimpleTestResult(false, getName()+": threw "+e);
			}
		}

		private static readonly String[] zipFileNames = {"ecbnk",
													"ecbnt",
													"ecbvk",
													"ecbvt"};

		public TestResult perform() {
			countOfTests = 0;
			for (int i = 0; i < zipFileNames.Length; i++) {
				TestResult res = performTestsFromZipFile(zipFileNames[i]);
				if (!res.isSuccessful()) {
					return res;
				}
			}
			return new SimpleTestResult(true, getName() + ": "+countOfTests+" performed Okay");
		}

		[STAThread]
		static void Main(string[] args)
		{
			AESVectorFileTest test = new AESVectorFileTest();
			TestResult result = test.perform();
			Console.WriteLine(result.ToString());

			test = new AESLightVectorFileTest();
			result = test.perform();
			Console.WriteLine(result.ToString());

			test = new AESFastVectorFileTest();
			result = test.perform();
			Console.WriteLine(result.ToString());

		}

		private class AESLightVectorFileTest : AESVectorFileTest {
			protected override BlockCipher createNewEngineForTest() {
				return new AESLightEngine();
			}

			public override String getName() {
				return "AESLight";
			}

		}

		private class AESFastVectorFileTest : AESVectorFileTest {
			protected override BlockCipher createNewEngineForTest() {
				return new AESFastEngine();
			}

			public override String getName() {
				return "AESFast";
			}

		}
	}
}
